@@ -1,7 +1,10 @@
+/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
+
 /* Metacity interface used by GTK+ UI to talk to core */
 
 /* 
  * Copyright (C) 2001 Havoc Pennington
+ * Copyright (C) 2005 Elijah Newren
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -31,6 +34,13 @@ void meta_core_get_client_size (Display *xdisplay,
                                 int     *width,
                                 int     *height);
 
+gboolean meta_core_titlebar_is_onscreen (Display *xdisplay,
+                                         Window   frame_xwindow);
+
+gboolean meta_core_window_has_frame (Display *xdisplay,
+                                     Window   frame_xwindow);
+
+
 Window meta_core_get_client_xwindow (Display *xdisplay,
                                      Window   frame_xwindow);
 
@@ -62,11 +72,11 @@ void meta_core_user_raise   (Display *xdisplay,
                              Window   frame_xwindow);
 void meta_core_user_lower_and_unfocus (Display *xdisplay,
                                        Window   frame_xwindow,
-                                       Time     timestamp);
+                                       guint32  timestamp);
 
 void meta_core_user_focus   (Display *xdisplay,
                              Window   frame_xwindow,
-                             Time     timestamp);
+                             guint32  timestamp);
 
 /* get position of client, same coord space expected by move */
 void meta_core_get_position (Display *xdisplay,
@@ -91,13 +101,19 @@ void meta_core_delete           (Display *xdisplay,
                                  Window   frame_xwindow,
                                  guint32  timestamp);
 void meta_core_unshade          (Display *xdisplay,
-                                 Window   frame_xwindow);
+                                 Window   frame_xwindow,
+                                 guint32  timestamp);
 void meta_core_shade            (Display *xdisplay,
-                                 Window   frame_xwindow);
+                                 Window   frame_xwindow,
+                                 guint32  timestamp);
 void meta_core_unstick          (Display *xdisplay,
                                  Window   frame_xwindow);
 void meta_core_stick            (Display *xdisplay,
                                  Window   frame_xwindow);
+void meta_core_unmake_above     (Display *xdisplay,
+                                 Window   frame_xwindow);
+void meta_core_make_above       (Display *xdisplay,
+                                 Window   frame_xwindow);
 void meta_core_change_workspace (Display *xdisplay,
                                  Window   frame_xwindow,
                                  int      new_workspace);
@@ -123,7 +139,7 @@ void meta_core_show_window_menu (Display *xdisplay,
                                  int      root_x,
                                  int      root_y,
                                  int      button,
-                                 Time     timestamp);
+                                 guint32  timestamp);
 
 void meta_core_get_menu_accelerator (MetaMenuOp           menu_op,
                                      int                  workspace,
@@ -137,11 +153,11 @@ gboolean   meta_core_begin_grab_op (Display    *xdisplay,
                                     int         event_serial,
                                     int         button,
                                     gulong      modmask,
-                                    Time        timestamp,
+                                    guint32     timestamp,
                                     int         root_x,
                                     int         root_y);
 void       meta_core_end_grab_op   (Display    *xdisplay,
-                                    Time        timestamp);
+                                    guint32     timestamp);
 MetaGrabOp meta_core_get_grab_op     (Display    *xdisplay);
 Window     meta_core_get_grab_frame  (Display   *xdisplay);
 int        meta_core_get_grab_button (Display  *xdisplay);
@@ -167,6 +183,8 @@ void meta_core_increment_event_serial (Display *display);
 
 int meta_ui_get_last_event_serial (Display *xdisplay);
 
+void meta_invalidate_default_icons (void);
+
 #endif
 
 
