@@ -1,3 +1,5 @@
+/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
+
 /* Metacity window frame manager widget */
 
 /* 
@@ -22,6 +24,8 @@
  */
 
 #include <config.h>
+#include <math.h>
+#include "boxes.h"
 #include "frames.h"
 #include "util.h"
 #include "core.h"
@@ -45,6 +49,9 @@ static void meta_frames_style_set  (GtkWidget       *widget,
 static void meta_frames_realize    (GtkWidget       *widget);
 static void meta_frames_unrealize  (GtkWidget       *widget);
 
+static void meta_frames_update_prelit_control (MetaFrames      *frames,
+                                               MetaUIFrame     *frame,
+                                               MetaFrameControl control);
 static gboolean meta_frames_button_press_event    (GtkWidget           *widget,
                                                    GdkEventButton      *event);
 static gboolean meta_frames_button_release_event  (GtkWidget           *widget,
@@ -232,21 +239,16 @@ meta_frames_destroy (GtkObject *object)
   clear_tip (frames);
   
   winlist = NULL;
-  g_hash_table_foreach (frames->frames,
-                        listify_func,
-                        &winlist);
+  g_hash_table_foreach (frames->frames, listify_func, &winlist);
 
   /* Unmanage all frames */
-  tmp = winlist;
-  while (tmp != NULL)
+  for (tmp = winlist; tmp != NULL; tmp = tmp->next)
     {
       MetaUIFrame *frame;
 
       frame = tmp->data;
 
       meta_frames_unmanage_window (frames, frame->xwindow);
-      
-      tmp = tmp->next;
     }
   g_slist_free (winlist);
 
@@ -277,18 +279,16 @@ meta_frames_finalize (GObject *object)
 
 typedef struct
 {
-  GdkPixmap *top;
-  gint top_x, top_y;
-  
-  GdkPixmap *left;
-  gint left_x, left_y;
-  
-  GdkPixmap *right;
-  gint right_y, right_x;
-  
-  GdkPixmap *bottom;
-  gint bottom_x, bottom_y;
-  
+  MetaRectangle rect;
+  GdkPixmap *pixmap;
+} CachedFramePiece;
+
+typedef struct
+{
+  /* Caches of the four rendered sides in a MetaFrame.
+   * Order: top (titlebar), left, right, bottom.
+   */
+  CachedFramePiece piece[4];
 } CachedPixels;
 
 static CachedPixels *
@@ -313,18 +313,13 @@ invalidate_cache (MetaFrames *frames,
 		  MetaUIFrame *frame)
 {
   CachedPixels *pixels = get_cache (frames, frame);
+  int i;
   
-  if (pixels->top)
-    g_object_unref (pixels->top);
-  if (pixels->left)
-    g_object_unref (pixels->left);
-  if (pixels->right)
-    g_object_unref (pixels->right);
-  if (pixels->bottom)
-    g_object_unref (pixels->bottom);
+  for (i = 0; i < 4; i++)
+    if (pixels->piece[i].pixmap)
+      g_object_unref (pixels->piece[i].pixmap);
   
   g_free (pixels);
-  
   g_hash_table_remove (frames->cache, frame);
 }
 
@@ -763,10 +758,10 @@ meta_frames_apply_shapes (MetaFrames *frames,
 
   meta_frames_calc_geometry (frames, frame, &fgeom);
 
-  if (!(fgeom.top_left_corner_rounded ||
-        fgeom.top_right_corner_rounded ||
-        fgeom.bottom_left_corner_rounded ||
-        fgeom.bottom_right_corner_rounded ||
+  if (!(fgeom.top_left_corner_rounded_radius != 0 ||
+        fgeom.top_right_corner_rounded_radius != 0 ||
+        fgeom.bottom_left_corner_rounded_radius != 0 ||
+        fgeom.bottom_right_corner_rounded_radius != 0 ||
         window_has_shape))
     {
       if (frame->shape_applied)
@@ -791,102 +786,72 @@ meta_frames_apply_shapes (MetaFrames *frames,
   
   corners_xregion = XCreateRegion ();
   
-  if (fgeom.top_left_corner_rounded)
+  if (fgeom.top_left_corner_rounded_radius != 0)
     {
-      xrect.x = 0;
-      xrect.y = 0;
-      xrect.width = 5;
-      xrect.height = 1;
-      
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      const int radius = fgeom.top_left_corner_rounded_radius;
+      int i;
 
-      xrect.y = 1;
-      xrect.width = 3;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
-      
-      xrect.y = 2;
-      xrect.width = 2;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
-
-      xrect.y = 3;
-      xrect.width = 1;
-      xrect.height = 2;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      for (i=0; i<radius; i++)
+        {
+          const int width = 1 + (radius - floor(sqrt(radius*radius - (radius-i)*(radius-i))));
+          xrect.x = 0;
+          xrect.y = i;
+          xrect.width = width;
+          xrect.height = 1;
+          
+          XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+        }
     }
 
-  if (fgeom.top_right_corner_rounded)
+  if (fgeom.top_right_corner_rounded_radius != 0)
     {
-      xrect.x = new_window_width - 5;
-      xrect.y = 0;
-      xrect.width = 5;
-      xrect.height = 1;
-      
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      const int radius = fgeom.top_right_corner_rounded_radius;
+      int i;
 
-      xrect.y = 1;
-      xrect.x = new_window_width - 3;
-      xrect.width = 3;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
-      
-      xrect.y = 2;
-      xrect.x = new_window_width - 2;
-      xrect.width = 2;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
-
-      xrect.y = 3;
-      xrect.x = new_window_width - 1;
-      xrect.width = 1;
-      xrect.height = 2;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      for (i=0; i<radius; i++)
+        {
+          const int width = 1 + (radius - floor(sqrt(radius*radius - (radius-i)*(radius-i))));
+          xrect.x = new_window_width - width;
+          xrect.y = i;
+          xrect.width = width;
+          xrect.height = 1;
+          
+          XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+        }
     }
 
-  if (fgeom.bottom_left_corner_rounded)
+  if (fgeom.bottom_left_corner_rounded_radius != 0)
     {
-      xrect.x = 0;
-      xrect.y = new_window_height - 1;
-      xrect.width = 5;
-      xrect.height = 1;
-      
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
-      
-      xrect.y = new_window_height - 2;
-      xrect.width = 3;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
-      
-      xrect.y = new_window_height - 3;
-      xrect.width = 2;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      const int radius = fgeom.bottom_left_corner_rounded_radius;
+      int i;
 
-      xrect.y = new_window_height - 5;
-      xrect.width = 1;
-      xrect.height = 2;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      for (i=0; i<radius; i++)
+        {
+          const int width = 1 + (radius - floor(sqrt(radius*radius - (radius-i)*(radius-i))));
+          xrect.x = 0;
+          xrect.y = new_window_height - i;
+          xrect.width = width;
+          xrect.height = 1;
+          
+          XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+        }
     }
 
-  if (fgeom.bottom_right_corner_rounded)
+  if (fgeom.bottom_right_corner_rounded_radius != 0)
     {
-      xrect.x = new_window_width - 5;
-      xrect.y = new_window_height - 1;
-      xrect.width = 5;
-      xrect.height = 1;
-      
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
-      
-      xrect.y = new_window_height - 2;
-      xrect.x = new_window_width - 3;
-      xrect.width = 3;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
-      
-      xrect.y = new_window_height - 3;
-      xrect.x = new_window_width - 2;
-      xrect.width = 2;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      const int radius = fgeom.bottom_right_corner_rounded_radius;
+      int i;
 
-      xrect.y = new_window_height - 5;
-      xrect.x = new_window_width - 1;
-      xrect.width = 1;
-      xrect.height = 2;
-      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      for (i=0; i<radius; i++)
+        {
+          const int width = 1 + (radius - floor(sqrt(radius*radius - (radius-i)*(radius-i))));
+          xrect.x = new_window_width - width;
+          xrect.y = new_window_height - i;
+          xrect.width = width;
+          xrect.height = 1;
+          
+          XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+        }
     }
   
   window_xregion = XCreateRegion ();
@@ -1108,6 +1073,24 @@ show_tip_now (MetaFrames *frames)
     case META_FRAME_CONTROL_UNMAXIMIZE:
       tiptext = _("Unmaximize Window");
       break;
+    case META_FRAME_CONTROL_SHADE:
+      tiptext = _("Roll Up Window");
+      break;
+    case META_FRAME_CONTROL_UNSHADE:
+      tiptext = _("Unroll Window");
+      break;
+    case META_FRAME_CONTROL_ABOVE:
+      tiptext = _("Keep Window On Top");
+      break;
+    case META_FRAME_CONTROL_UNABOVE:
+      tiptext = _("Remove Window From Top");
+      break;
+    case META_FRAME_CONTROL_STICK:
+      tiptext = _("Always On Visible Workspace");
+      break;
+    case META_FRAME_CONTROL_UNSTICK:
+      tiptext = _("Put Window On Only One Workspace");
+      break;
     case META_FRAME_CONTROL_RESIZE_SE:
       break;
     case META_FRAME_CONTROL_RESIZE_S:
@@ -1204,15 +1187,6 @@ redraw_control (MetaFrames *frames,
   invalidate_cache (frames, frame);
 }
 
-static gboolean
-point_in_control (MetaFrames *frames,
-                  MetaUIFrame *frame,
-                  MetaFrameControl control,
-                  int x, int y)
-{
-  return control == get_control (frames, frame, x, y);
-}
-
 static gboolean
 meta_frames_button_press_event (GtkWidget      *widget,
                                 GdkEventButton *event)
@@ -1271,10 +1245,12 @@ meta_frames_button_press_event (GtkWidget      *widget,
               {
                 if (flags & META_FRAME_SHADED)
                   meta_core_unshade (gdk_display,
-                                     frame->xwindow);
+                                     frame->xwindow,
+                                     event->time);
                 else
                   meta_core_shade (gdk_display,
-                                   frame->xwindow);
+                                   frame->xwindow,
+                                   event->time);
               }
           }
           break;          
@@ -1292,9 +1268,14 @@ meta_frames_button_press_event (GtkWidget      *widget,
 
         case META_ACTION_DOUBLE_CLICK_TITLEBAR_MINIMIZE:
           {
-            meta_core_minimize (gdk_display, frame->xwindow);
-            break;
+            flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
+            
+            if (flags & META_FRAME_ALLOWS_MINIMIZE)
+              {
+                meta_core_minimize (gdk_display, frame->xwindow);
+              }
           }
+          break;
 
         case META_ACTION_DOUBLE_CLICK_TITLEBAR_NONE:
           /* Yaay, a sane user that doesn't use that other weird crap! */
@@ -1316,6 +1297,12 @@ meta_frames_button_press_event (GtkWidget      *widget,
        control == META_FRAME_CONTROL_UNMAXIMIZE ||
        control == META_FRAME_CONTROL_MINIMIZE ||
        control == META_FRAME_CONTROL_DELETE ||
+       control == META_FRAME_CONTROL_SHADE ||
+       control == META_FRAME_CONTROL_UNSHADE ||
+       control == META_FRAME_CONTROL_ABOVE ||
+       control == META_FRAME_CONTROL_UNABOVE ||
+       control == META_FRAME_CONTROL_STICK ||
+       control == META_FRAME_CONTROL_UNSTICK ||
        control == META_FRAME_CONTROL_MENU))
     {
       MetaGrabOp op = META_GRAB_OP_NONE;
@@ -1337,6 +1324,24 @@ meta_frames_button_press_event (GtkWidget      *widget,
         case META_FRAME_CONTROL_MENU:
           op = META_GRAB_OP_CLICKING_MENU;
           break;
+        case META_FRAME_CONTROL_SHADE:
+          op = META_GRAB_OP_CLICKING_SHADE;
+          break;
+        case META_FRAME_CONTROL_UNSHADE:
+          op = META_GRAB_OP_CLICKING_UNSHADE;
+          break;
+        case META_FRAME_CONTROL_ABOVE:
+          op = META_GRAB_OP_CLICKING_ABOVE;
+          break;
+        case META_FRAME_CONTROL_UNABOVE:
+          op = META_GRAB_OP_CLICKING_UNABOVE;
+          break;
+        case META_FRAME_CONTROL_STICK:
+          op = META_GRAB_OP_CLICKING_STICK;
+          break;
+        case META_FRAME_CONTROL_UNSTICK:
+          op = META_GRAB_OP_CLICKING_UNSTICK;
+          break;
         default:
           g_assert_not_reached ();
           break;
@@ -1353,6 +1358,7 @@ meta_frames_button_press_event (GtkWidget      *widget,
                                event->x_root,
                                event->y_root);      
       
+      frame->prelit_control = control;
       redraw_control (frames, frame, control);
 
       if (op == META_GRAB_OP_CLICKING_MENU)
@@ -1536,67 +1542,96 @@ meta_frames_button_release_event    (GtkWidget           *widget,
   if (frame->xwindow == meta_core_get_grab_frame (gdk_display) &&
       ((int) event->button) == meta_core_get_grab_button (gdk_display))
     {
-      gboolean end_grab;
-
-      end_grab = FALSE;
+      MetaFrameControl control;
 
+      control = get_control (frames, frame, event->x, event->y);
+      
       switch (op)
         {
         case META_GRAB_OP_CLICKING_MINIMIZE:
-          if (point_in_control (frames, frame,
-                                META_FRAME_CONTROL_MINIMIZE,
-                                event->x, event->y))
+          if (control == META_FRAME_CONTROL_MINIMIZE)
             meta_core_minimize (gdk_display, frame->xwindow);
-
-          redraw_control (frames, frame,
-                          META_FRAME_CONTROL_MINIMIZE);
-          end_grab = TRUE;
+          
+          meta_core_end_grab_op (gdk_display, event->time);
           break;
 
         case META_GRAB_OP_CLICKING_MAXIMIZE:
-          if (point_in_control (frames, frame,
-                                META_FRAME_CONTROL_MAXIMIZE,
-                                event->x, event->y))
+          if (control == META_FRAME_CONTROL_MAXIMIZE)
             meta_core_maximize (gdk_display, frame->xwindow);
-
-          redraw_control (frames, frame,
-                          META_FRAME_CONTROL_MAXIMIZE);
-          end_grab = TRUE;
+          
+          meta_core_end_grab_op (gdk_display, event->time);
           break;
 
         case META_GRAB_OP_CLICKING_UNMAXIMIZE:
-          if (point_in_control (frames, frame,
-                                META_FRAME_CONTROL_UNMAXIMIZE,
-                                event->x, event->y))
+          if (control == META_FRAME_CONTROL_UNMAXIMIZE)
             meta_core_unmaximize (gdk_display, frame->xwindow);
-
-          redraw_control (frames, frame,
-                          META_FRAME_CONTROL_MAXIMIZE);
-          end_grab = TRUE;
+          
+          meta_core_end_grab_op (gdk_display, event->time);
           break;
           
         case META_GRAB_OP_CLICKING_DELETE:
-          if (point_in_control (frames, frame,
-                                META_FRAME_CONTROL_DELETE,
-                                event->x, event->y))
+          if (control == META_FRAME_CONTROL_DELETE)
             meta_core_delete (gdk_display, frame->xwindow, event->time);
-          redraw_control (frames, frame,
-                          META_FRAME_CONTROL_DELETE);
-          end_grab = TRUE;
+          
+          meta_core_end_grab_op (gdk_display, event->time);
           break;
           
         case META_GRAB_OP_CLICKING_MENU:
-          redraw_control (frames, frame,
-                          META_FRAME_CONTROL_MENU);
-          end_grab = TRUE;
+          meta_core_end_grab_op (gdk_display, event->time);
           break;
 
+        case META_GRAB_OP_CLICKING_SHADE:
+          if (control == META_FRAME_CONTROL_SHADE)
+            meta_core_shade (gdk_display, frame->xwindow, event->time);
+          
+          meta_core_end_grab_op (gdk_display, event->time);
+          break;
+ 
+        case META_GRAB_OP_CLICKING_UNSHADE:
+          if (control == META_FRAME_CONTROL_UNSHADE)
+            meta_core_unshade (gdk_display, frame->xwindow, event->time);
+
+          meta_core_end_grab_op (gdk_display, event->time);
+          break;
+
+        case META_GRAB_OP_CLICKING_ABOVE:
+          if (control == META_FRAME_CONTROL_ABOVE)
+            meta_core_make_above (gdk_display, frame->xwindow);
+          
+          meta_core_end_grab_op (gdk_display, event->time);
+          break;
+ 
+        case META_GRAB_OP_CLICKING_UNABOVE:
+          if (control == META_FRAME_CONTROL_UNABOVE)
+            meta_core_unmake_above (gdk_display, frame->xwindow);
+
+          meta_core_end_grab_op (gdk_display, event->time);
+          break;
+
+        case META_GRAB_OP_CLICKING_STICK:
+          if (control == META_FRAME_CONTROL_STICK)
+            meta_core_stick (gdk_display, frame->xwindow);
+
+          meta_core_end_grab_op (gdk_display, event->time);
+          break;
+ 
+        case META_GRAB_OP_CLICKING_UNSTICK:
+          if (control == META_FRAME_CONTROL_UNSTICK)
+            meta_core_unstick (gdk_display, frame->xwindow);
+
+          meta_core_end_grab_op (gdk_display, event->time);
+          break;
+          
         default:
           break;
         }
 
-      if (end_grab)
-        meta_core_end_grab_op (gdk_display, event->time);
+      /* Update the prelit control regardless of what button the mouse
+       * was released over; needed so that the new button can become
+       * prelit so to let the user know that it can now be pressed.
+       * :)
+       */
+      meta_frames_update_prelit_control (frames, frame, control);
     }
   
   return TRUE;
@@ -1610,6 +1645,7 @@ meta_frames_update_prelit_control (MetaFrames      *frames,
   MetaFrameControl old_control;
   MetaCursor cursor;
 
+
   meta_verbose ("Updating prelit control from %u to %u\n",
                 frame->prelit_control, control);
   
@@ -1633,6 +1669,18 @@ meta_frames_update_prelit_control (MetaFrames      *frames,
       break;
     case META_FRAME_CONTROL_UNMAXIMIZE:
       break;
+    case META_FRAME_CONTROL_SHADE:
+      break;
+    case META_FRAME_CONTROL_UNSHADE:
+      break;
+    case META_FRAME_CONTROL_ABOVE:
+      break;
+    case META_FRAME_CONTROL_UNABOVE:
+      break;
+    case META_FRAME_CONTROL_STICK:
+      break;
+    case META_FRAME_CONTROL_UNSTICK:
+      break;
     case META_FRAME_CONTROL_RESIZE_SE:
       cursor = META_CURSOR_SE_RESIZE;
       break;
@@ -1670,6 +1718,12 @@ meta_frames_update_prelit_control (MetaFrames      *frames,
     case META_FRAME_CONTROL_MINIMIZE:
     case META_FRAME_CONTROL_MAXIMIZE:
     case META_FRAME_CONTROL_DELETE:
+    case META_FRAME_CONTROL_SHADE:
+    case META_FRAME_CONTROL_UNSHADE:
+    case META_FRAME_CONTROL_ABOVE:
+    case META_FRAME_CONTROL_UNABOVE:
+    case META_FRAME_CONTROL_STICK:
+    case META_FRAME_CONTROL_UNSTICK:
     case META_FRAME_CONTROL_UNMAXIMIZE:
       /* leave control set */
       break;
@@ -1718,6 +1772,12 @@ meta_frames_motion_notify_event     (GtkWidget           *widget,
     case META_GRAB_OP_CLICKING_MINIMIZE:
     case META_GRAB_OP_CLICKING_MAXIMIZE:
     case META_GRAB_OP_CLICKING_UNMAXIMIZE:
+    case META_GRAB_OP_CLICKING_SHADE:
+    case META_GRAB_OP_CLICKING_UNSHADE:
+    case META_GRAB_OP_CLICKING_ABOVE:
+    case META_GRAB_OP_CLICKING_UNABOVE:
+    case META_GRAB_OP_CLICKING_STICK:
+    case META_GRAB_OP_CLICKING_UNSTICK:
       {
         MetaFrameControl control;
         int x, y;
@@ -1736,8 +1796,20 @@ meta_frames_motion_notify_event     (GtkWidget           *widget,
                 grab_op == META_GRAB_OP_CLICKING_MINIMIZE) ||
                (control == META_FRAME_CONTROL_MAXIMIZE &&
                 (grab_op == META_GRAB_OP_CLICKING_MAXIMIZE ||
-                 grab_op == META_GRAB_OP_CLICKING_UNMAXIMIZE))))
-          control = META_FRAME_CONTROL_NONE;
+                 grab_op == META_GRAB_OP_CLICKING_UNMAXIMIZE)) ||
+               (control == META_FRAME_CONTROL_SHADE &&
+                grab_op == META_GRAB_OP_CLICKING_SHADE) ||
+               (control == META_FRAME_CONTROL_UNSHADE &&
+                grab_op == META_GRAB_OP_CLICKING_UNSHADE) ||
+               (control == META_FRAME_CONTROL_ABOVE &&
+                grab_op == META_GRAB_OP_CLICKING_ABOVE) ||
+               (control == META_FRAME_CONTROL_UNABOVE &&
+                grab_op == META_GRAB_OP_CLICKING_UNABOVE) ||
+               (control == META_FRAME_CONTROL_STICK &&
+                grab_op == META_GRAB_OP_CLICKING_STICK) ||
+               (control == META_FRAME_CONTROL_UNSTICK &&
+                grab_op == META_GRAB_OP_CLICKING_UNSTICK)))
+           control = META_FRAME_CONTROL_NONE;
         
         /* Update prelit control and cursor */
 	meta_frames_update_prelit_control (frames, frame, control);
@@ -1834,24 +1906,25 @@ clear_backing (GdkPixmap *pixmap,
   g_object_unref (tmp_gc);
 }
 
+/* Returns a pixmap with a piece of the windows frame painted on it.
+*/
+
 static GdkPixmap *
 generate_pixmap (MetaFrames *frames,
 		 MetaUIFrame *frame,
-		 int x,
-		 int y,
-		 int width,
-		 int height)
+                 MetaRectangle rect)
 {
   GdkRectangle rectangle;
   GdkRegion *region;
   GdkPixmap *result;
 
-  rectangle.x = x;
-  rectangle.y = y;
-  rectangle.width = MAX (width, 1);
-  rectangle.height = MAX (height, 1);
+  rectangle.x = rect.x;
+  rectangle.y = rect.y;
+  rectangle.width = MAX (rect.width, 1);
+  rectangle.height = MAX (rect.height, 1);
   
-  result = gdk_pixmap_new (frame->window, rectangle.width, rectangle.height, -1);
+  result = gdk_pixmap_new (frame->window,
+                           rectangle.width, rectangle.height, -1);
   
   clear_backing (result, frame->window, rectangle.x, rectangle.y);
 
@@ -1865,6 +1938,7 @@ generate_pixmap (MetaFrames *frames,
   return result;
 }
 
+
 static void
 populate_cache (MetaFrames *frames,
 		MetaUIFrame *frame)
@@ -1873,6 +1947,7 @@ populate_cache (MetaFrames *frames,
   int width, height;
   int frame_width, frame_height, screen_width, screen_height;
   CachedPixels *pixels;
+  int i;
 
   meta_core_get_frame_extents (gdk_display,
                                frame->xwindow,
@@ -1899,37 +1974,34 @@ populate_cache (MetaFrames *frames,
   meta_core_get_client_size (gdk_display, frame->xwindow, &width, &height);
 
   pixels = get_cache (frames, frame);
-  
-  if (!pixels->top)
+
+  /* Setup the rectangles for the four frame borders. First top, then
+     left, right and bottom. */
+  pixels->piece[0].rect.x = 0;
+  pixels->piece[0].rect.y = 0;
+  pixels->piece[0].rect.width = left + width + right;
+  pixels->piece[0].rect.height = top;
+
+  pixels->piece[1].rect.x = 0;
+  pixels->piece[1].rect.y = top;
+  pixels->piece[1].rect.width = left;
+  pixels->piece[1].rect.height = height;
+
+  pixels->piece[2].rect.x = left + width;
+  pixels->piece[2].rect.y = top;
+  pixels->piece[2].rect.width = right;
+  pixels->piece[2].rect.height = height;
+
+  pixels->piece[3].rect.x = 0;
+  pixels->piece[3].rect.y = top + height;
+  pixels->piece[3].rect.width = left + width + right;
+  pixels->piece[3].rect.height = bottom;
+
+  for (i = 0; i < 4; i++)
     {
-      pixels->top = generate_pixmap (frames, frame,
-				    0, 0, left + width + height, top);
-      pixels->top_x = 0;
-      pixels->top_y = 0;
-    }
-  
-  if (!pixels->left)
-    {
-      pixels->left = generate_pixmap (frames, frame,
-				     0, top, left, height);
-      pixels->left_x = 0;
-      pixels->left_y = top;
-    }
-  
-  if (!pixels->right)
-    {
-      pixels->right = generate_pixmap (frames, frame,
-				      left + width, top, right, height);
-      pixels->right_x = left + width;
-      pixels->right_y = top;
-    }
-  
-  if (!pixels->bottom)
-    {
-      pixels->bottom = generate_pixmap (frames, frame,
-				       0, top + height, left + width + height, bottom);
-      pixels->bottom_x = 0;
-      pixels->bottom_y = top + height;
+      CachedFramePiece *piece = &pixels->piece[i];
+      if (!piece->pixmap)
+        piece->pixmap = generate_pixmap (frames, frame, piece->rect);
     }
   
   if (frames->invalidate_cache_timeout_id)
@@ -1987,13 +2059,41 @@ subtract_from_region (GdkRegion *region, GdkDrawable *drawable,
   gdk_region_destroy (reg_rect);
 }
 
+static void
+cached_pixels_draw (CachedPixels *pixels,
+                    GdkWindow *window,
+                    GdkRegion *region)
+{
+  GdkGC *gc;
+  int i;
+
+  gc = gdk_gc_new (window);
+
+  for (i = 0; i < 4; i++)
+    {
+      CachedFramePiece *piece;
+      piece = &pixels->piece[i];
+      
+      if (piece->pixmap)
+        {
+          gdk_draw_drawable (window, gc, piece->pixmap,
+                             0, 0,
+                             piece->rect.x, piece->rect.y,
+                             -1, -1);
+          subtract_from_region (region, piece->pixmap,
+          piece->rect.x, piece->rect.y);
+        }
+    }
+  
+  g_object_unref (gc);
+}
+
 static gboolean
 meta_frames_expose_event (GtkWidget           *widget,
                           GdkEventExpose      *event)
 {
   MetaUIFrame *frame;
   MetaFrames *frames;
-  GdkGC *gc;
   GdkRegion *region;
   CachedPixels *pixels;
 
@@ -2014,51 +2114,13 @@ meta_frames_expose_event (GtkWidget           *widget,
 
   region = gdk_region_copy (event->region);
   
-  gc = gdk_gc_new (frame->window);
-
   pixels = get_cache (frames, frame);
+
+  cached_pixels_draw (pixels, frame->window, region);
   
-  if (pixels->top)
-    {
-      gdk_draw_drawable (frame->window, gc, pixels->top,
-			 0, 0,
-			 pixels->top_x, pixels->top_y,
-			 -1, -1);
-      subtract_from_region (region, pixels->top, pixels->top_x, pixels->top_y);
-    }
-
-  if (pixels->left)
-    {
-      gdk_draw_drawable (frame->window, gc, pixels->left,
-			 0, 0,
-			 pixels->left_x, pixels->left_y,
-			 -1, -1);
-      subtract_from_region (region, pixels->left, pixels->left_x, pixels->left_y);
-    }
-
-  if (pixels->right)
-    {
-      gdk_draw_drawable (frame->window, gc, pixels->right,
-			 0, 0,
-			 pixels->right_x, pixels->right_y,
-			 -1, -1);
-      subtract_from_region (region, pixels->right, pixels->right_x, pixels->right_y);
-    }
-  
-  if (pixels->bottom)
-    {
-      gdk_draw_drawable (frame->window, gc, pixels->bottom,
-			 0, 0,
-			 pixels->bottom_x, pixels->bottom_y,
-			 -1, -1);
-      subtract_from_region (region, pixels->bottom, pixels->bottom_x, pixels->bottom_y);
-    }
-
   clip_to_screen (region, frame);
-  meta_frames_paint_to_drawable (frames, frame,
-				 frame->window, region, 0, 0);
-  
-  g_object_unref (gc);
+  meta_frames_paint_to_drawable (frames, frame, frame->window, region, 0, 0);
+
   gdk_region_destroy (region);
   
   return TRUE;
@@ -2092,13 +2154,8 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
   
   widget = GTK_WIDGET (frames);
 
-  i = 0;
-  while (i < META_BUTTON_TYPE_LAST)
-    {
-      button_states[i] = META_BUTTON_STATE_NORMAL;
-      
-      ++i;
-    }
+  for (i = 0; i < META_BUTTON_TYPE_LAST; i++)
+    button_states[i] = META_BUTTON_STATE_NORMAL;
 
   grab_frame = meta_core_get_grab_frame (gdk_display);
   grab_op = meta_core_get_grab_op (gdk_display);
@@ -2132,6 +2189,42 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
       else
         button_states[META_BUTTON_TYPE_MAXIMIZE] = META_BUTTON_STATE_PRELIGHT;
       break;
+    case META_FRAME_CONTROL_SHADE:
+      if (grab_op == META_GRAB_OP_CLICKING_SHADE)
+        button_states[META_BUTTON_TYPE_SHADE] = META_BUTTON_STATE_PRESSED;
+      else
+        button_states[META_BUTTON_TYPE_SHADE] = META_BUTTON_STATE_PRELIGHT;
+      break;
+    case META_FRAME_CONTROL_UNSHADE:
+      if (grab_op == META_GRAB_OP_CLICKING_UNSHADE)
+        button_states[META_BUTTON_TYPE_UNSHADE] = META_BUTTON_STATE_PRESSED;
+      else
+        button_states[META_BUTTON_TYPE_UNSHADE] = META_BUTTON_STATE_PRELIGHT;
+      break;
+    case META_FRAME_CONTROL_ABOVE:
+      if (grab_op == META_GRAB_OP_CLICKING_ABOVE)
+        button_states[META_BUTTON_TYPE_ABOVE] = META_BUTTON_STATE_PRESSED;
+      else
+        button_states[META_BUTTON_TYPE_ABOVE] = META_BUTTON_STATE_PRELIGHT;
+      break;
+    case META_FRAME_CONTROL_UNABOVE:
+      if (grab_op == META_GRAB_OP_CLICKING_UNABOVE)
+        button_states[META_BUTTON_TYPE_UNABOVE] = META_BUTTON_STATE_PRESSED;
+      else
+        button_states[META_BUTTON_TYPE_UNABOVE] = META_BUTTON_STATE_PRELIGHT;
+      break;
+    case META_FRAME_CONTROL_STICK:
+      if (grab_op == META_GRAB_OP_CLICKING_STICK)
+        button_states[META_BUTTON_TYPE_STICK] = META_BUTTON_STATE_PRESSED;
+      else
+        button_states[META_BUTTON_TYPE_STICK] = META_BUTTON_STATE_PRELIGHT;
+      break;
+    case META_FRAME_CONTROL_UNSTICK:
+      if (grab_op == META_GRAB_OP_CLICKING_UNSTICK)
+        button_states[META_BUTTON_TYPE_UNSTICK] = META_BUTTON_STATE_PRESSED;
+      else
+        button_states[META_BUTTON_TYPE_UNSTICK] = META_BUTTON_STATE_PRELIGHT;
+      break;
     case META_FRAME_CONTROL_DELETE:
       if (grab_op == META_GRAB_OP_CLICKING_DELETE)
         button_states[META_BUTTON_TYPE_CLOSE] = META_BUTTON_STATE_PRESSED;
@@ -2187,13 +2280,10 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
   gdk_region_get_rectangles (edges, &areas, &n_areas);
 
   meta_prefs_get_button_layout (&button_layout);
-  
-  i = 0;
-  while (i < n_areas)
+  for (i = 0; i < n_areas; i++)
     {
       if (GDK_IS_WINDOW (drawable))
         gdk_window_begin_paint_rect (drawable, &areas[i]);
-      
       meta_theme_draw_frame (meta_theme_get_current (),
                              widget,
                              drawable,
@@ -2210,8 +2300,6 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
 
       if (GDK_IS_WINDOW (drawable))
         gdk_window_end_paint (drawable);
-      
-      ++i;
     }
 
   gdk_region_destroy (edges);
@@ -2222,17 +2310,52 @@ static void
 meta_frames_set_window_background (MetaFrames   *frames,
                                    MetaUIFrame  *frame)
 {
-  gtk_style_set_background (GTK_WIDGET (frames)->style,
-                            frame->window, GTK_STATE_NORMAL);
+  MetaFrameFlags flags;
+  MetaFrameType type;
+  MetaFrameStyle *style;
+  gboolean frame_exists;
 
-#if 0
-  /* This is what we want for transparent background */
-  {
-    col.pixel = 0;
-    gdk_window_set_background (window, &col);
-  }
-#endif
-}
+  frame_exists = meta_core_window_has_frame (gdk_display, frame->xwindow);
+
+  if (frame_exists)
+    {
+      flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
+      type = meta_core_get_frame_type (gdk_display, frame->xwindow);
+      style = meta_theme_get_frame_style (meta_theme_get_current (),
+                                          type, flags);
+    }
+
+  if (frame_exists && style->window_background_color != NULL)
+    {
+      GdkColor color;
+      GdkVisual *visual;
+
+      meta_color_spec_render (style->window_background_color,
+                              GTK_WIDGET (frames),
+                              &color);
+
+      /* Fill in color.pixel */
+
+      gdk_rgb_find_color (gtk_widget_get_colormap (GTK_WIDGET (frames)),
+                          &color);
+
+      /* Set A in ARGB to window_background_alpha, if we have ARGB */
+
+      visual = gtk_widget_get_visual (GTK_WIDGET (frames));
+      if (visual->depth == 32) /* we have ARGB */
+        {
+          color.pixel = (color.pixel & 0xffffff) &
+            style->window_background_alpha << 24;
+        }
+
+      gdk_window_set_background (frame->window, &color);
+    }
+  else
+    {
+      gtk_style_set_background (GTK_WIDGET (frames)->style,
+                                frame->window, GTK_STATE_NORMAL);
+    }
+ }
 
 static gboolean
 meta_frames_enter_notify_event      (GtkWidget           *widget,
@@ -2287,17 +2410,35 @@ control_rect (MetaFrameControl control,
       rect = &fgeom->title_rect;
       break;
     case META_FRAME_CONTROL_DELETE:
-      rect = &fgeom->close_rect;
+      rect = &fgeom->close_rect.visible;
       break;
     case META_FRAME_CONTROL_MENU:
-      rect = &fgeom->menu_rect;
+      rect = &fgeom->menu_rect.visible;
       break;
     case META_FRAME_CONTROL_MINIMIZE:
-      rect = &fgeom->min_rect;
+      rect = &fgeom->min_rect.visible;
       break;
     case META_FRAME_CONTROL_MAXIMIZE:
     case META_FRAME_CONTROL_UNMAXIMIZE:
-      rect = &fgeom->max_rect;
+      rect = &fgeom->max_rect.visible;
+      break;
+    case META_FRAME_CONTROL_SHADE:
+      rect = &fgeom->shade_rect.visible;
+      break;
+    case META_FRAME_CONTROL_UNSHADE:
+      rect = &fgeom->unshade_rect.visible;
+      break;
+    case META_FRAME_CONTROL_ABOVE:
+      rect = &fgeom->above_rect.visible;
+      break;
+    case META_FRAME_CONTROL_UNABOVE:
+      rect = &fgeom->unabove_rect.visible;
+      break;
+    case META_FRAME_CONTROL_STICK:
+      rect = &fgeom->stick_rect.visible;
+      break;
+    case META_FRAME_CONTROL_UNSTICK:
+      rect = &fgeom->unstick_rect.visible;
       break;
     case META_FRAME_CONTROL_RESIZE_SE:
       break;
@@ -2335,7 +2476,6 @@ get_control (MetaFrames *frames,
   MetaFrameFlags flags;
   gboolean has_vert, has_horiz;
   GdkRectangle client;
-  int bottom_of_titlebar;
   
   meta_frames_calc_geometry (frames, frame, &fgeom);
 
@@ -2344,18 +2484,16 @@ get_control (MetaFrames *frames,
   client.width = fgeom.width - fgeom.left_width - fgeom.right_width;
   client.height = fgeom.height - fgeom.top_height - fgeom.bottom_height;  
 
-  bottom_of_titlebar = fgeom.title_rect.y + fgeom.title_rect.height;
-  
   if (POINT_IN_RECT (x, y, client))
     return META_FRAME_CONTROL_CLIENT_AREA;
   
-  if (POINT_IN_RECT (x, y, fgeom.close_rect))
+  if (POINT_IN_RECT (x, y, fgeom.close_rect.clickable))
     return META_FRAME_CONTROL_DELETE;
 
-  if (POINT_IN_RECT (x, y, fgeom.min_rect))
+  if (POINT_IN_RECT (x, y, fgeom.min_rect.clickable))
     return META_FRAME_CONTROL_MINIMIZE;
 
-  if (POINT_IN_RECT (x, y, fgeom.menu_rect))
+  if (POINT_IN_RECT (x, y, fgeom.menu_rect.clickable))
     return META_FRAME_CONTROL_MENU;
 
   flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
@@ -2371,7 +2509,7 @@ get_control (MetaFrames *frames,
         return META_FRAME_CONTROL_TITLE;
     }
 
-  if (POINT_IN_RECT (x, y, fgeom.max_rect))
+  if (POINT_IN_RECT (x, y, fgeom.max_rect.clickable))
     {
       if (flags & META_FRAME_MAXIMIZED)
         return META_FRAME_CONTROL_UNMAXIMIZE;
@@ -2379,6 +2517,36 @@ get_control (MetaFrames *frames,
         return META_FRAME_CONTROL_MAXIMIZE;
     }
       
+  if (POINT_IN_RECT (x, y, fgeom.shade_rect.clickable))
+    {
+      return META_FRAME_CONTROL_SHADE;
+    }
+
+  if (POINT_IN_RECT (x, y, fgeom.unshade_rect.clickable))
+    {
+      return META_FRAME_CONTROL_UNSHADE;
+    }
+
+  if (POINT_IN_RECT (x, y, fgeom.above_rect.clickable))
+    {
+      return META_FRAME_CONTROL_ABOVE;
+    }
+
+  if (POINT_IN_RECT (x, y, fgeom.unabove_rect.clickable))
+    {
+      return META_FRAME_CONTROL_UNABOVE;
+    }
+
+  if (POINT_IN_RECT (x, y, fgeom.stick_rect.clickable))
+    {
+      return META_FRAME_CONTROL_STICK;
+    }
+
+  if (POINT_IN_RECT (x, y, fgeom.unstick_rect.clickable))
+    {
+      return META_FRAME_CONTROL_UNSTICK;
+    }
+
   /* South resize always has priority over north resize,
    * in case of overlap.
    */
@@ -2446,7 +2614,7 @@ get_control (MetaFrames *frames,
         return META_FRAME_CONTROL_RESIZE_E;
     }
 
-  if (y >= bottom_of_titlebar)
+  if (y >= fgeom.top_height)
     return META_FRAME_CONTROL_NONE;
   else
     return META_FRAME_CONTROL_TITLE;
